# -*- coding: utf-8 -*-
import os
import sys
import inspect
from functools import partial, total_ordering

from yurlungur.core.wrapper import (
    YException, _YObject, _YAttr, OM
)
from yurlungur.tool.math import (
    YVector, YColor, YMatrix
)
from yurlungur.tool.meta import meta
from yurlungur.tool.util import trace


class YObject(_YObject):
    """base class
    >>> obj = YObject("pCone")
    >>> obj("cone")
    """

    def __init__(self, item):
        self.item = item

    def __repr__(self):
        if hasattr(meta, "SDNode"):
            return "id: " + self.name
        else:
            return self.name

    @property
    def name(self):
        if hasattr(meta, "SDNode"):
            return self.id
        else:
            return self.item

    @property
    def id(self):
        if hasattr(meta, "SDNode"):
            node_id = ""
            for node in meta.graph.getNodes():
                d = node.getDefinition()
                if d.getId() == self.item or d.getLabel() == self.item or node.getIdentifier() == self.item:
                    node_id = node.getIdentifier()
                    break
            return node_id if node_id else meta.graph.getIdentifier()

        if hasattr(meta, "ls"):
            return meta.ls(self.name, uuid=1)[0] or 0

        if hasattr(meta, "hda"):
            return meta.node(self.name).sessionId() or 0

        if hasattr(meta, "runtime"):
            return meta.runtime.getnodebyname(self.name).gbufferChannel or 0

        if hasattr(meta, "data"):
            return meta.data.objects[self.name].id_data or 0

        if hasattr(meta, "knob"):
            return meta.toNode(self.name)["name"].value() or 0

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.name).ID or 0

        if hasattr(meta, 'uclass'):
            return

        raise YException

    @trace
    def __call__(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            meta.graph.setIdentifier(args[0])

        if hasattr(meta, "rename"):
            return meta.rename(self.item, *args, **kwargs)

        if hasattr(meta, "hda"):
            return meta.node(self.item).setName(*args, **kwargs)

        if hasattr(meta, "runtime"):
            meta.runtime.getnodebyname(self.name).name = args[0]
            return YNode(args[0])

        if hasattr(meta, "data"):
            meta.data.objects[self.item].name = "".join(args)
            return "".join(args)

        if hasattr(meta, "knob"):
            meta.toNode(self.item).setName(args[0], **kwargs)

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.item).SetAttrs({"TOOLS_Name": args[0]})

        if hasattr(meta, 'uclass'):
            if meta.assets.rename_asset(self.name, args[0]):
                return YNode(args[0])

    @trace
    def __getattr__(self, val):
        if hasattr(meta, "SDNode"):
            prop = meta.graph.getNodeFromId(self.name).getPropertyFromId(val, meta.SDPropertyCategory.Input)
            return YAttr(meta.graph.getNodeFromId(self.name).getPropertyValue(prop).get(), self.name, val)

        if hasattr(meta, "getAttr"):
            return YAttr(meta.getAttr(self.name + "." + val), self.name, val)

        if hasattr(meta, "hda"):
            parm = (meta.node(self.name).parm(val) or meta.node(self.name).parmTuple(val))
            return YAttr(parm.eval(), self.name, val)

        if hasattr(meta, "runtime"):
            if '.' in self.name:
                node, prop = self.name.split('.')
                return YAttr(getattr(meta.runtime.getnodebyname(node), prop),
                             meta.runtime.getnodebyname(node).name, val)
            else:
                return YAttr(getattr(meta.runtime.getnodebyname(self.name), val), self.name, val)

        if hasattr(meta, "data"):
            return YAttr(meta.data.objects[self.name].name, self.name, val)

        if hasattr(meta, "knob"):
            return YAttr(meta.toNode(self.name)[val], self.name, val)

        if hasattr(meta, "fusion"):
            return YAttr(getattr(meta.fusion.GetCurrentComp().FindTool(self.name), val), self.name, val)

        if hasattr(meta, 'uclass'):
            return YAttr(
                meta.editor.get_actor_reference(self.name).get_editor_property(val), self.name, val)

        raise YException

    @trace
    def attr(self, val, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            prop = meta.graph.getNodeFromId(self.name).getPropertyFromId(val, meta.SDPropertyCategory.Input)
            return YAttr(meta.graph.getNodeFromId(self.name).getPropertyValue(prop).get(), self.name, val)

        if hasattr(meta, "getAttr"):
            return YAttr(meta.getAttr(self.name + "." + val, *args, **kwargs), self.name, val)

        if hasattr(meta, "hda"):
            parm = (meta.node(self.name).parm(val) or meta.node(self.name).parmTuple(val))
            return YAttr(parm.eval(), self.name, val)

        if hasattr(meta, "runtime"):
            return YAttr(getattr(meta.runtime.getnodebyname(self.name), val), self.name, val)

        if hasattr(meta, "data"):
            return YAttr(meta.data.objects[self.name].name, self.name, val)

        if hasattr(meta, "knob"):
            return YAttr(meta.toNode(self.name)[val], self.name, val)

        if hasattr(meta, "fusion"):
            return YAttr(getattr(meta.fusion.GetCurrentComp().FindTool(self.name), val), self.name, val)

        if hasattr(meta, 'uclass'):
            return YAttr(
                meta.editor.get_actor_reference(self.name).get_editor_property(val), self.name, val)

        raise YException

    @property
    def attrs(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            return tuple([
                prop.getId() for prop in
                meta.graph.getNodeFromId(self.name).getProperties(meta.SDPropertyCategory.Input)
            ])

        if hasattr(meta, "listAttr"):
            return tuple(meta.listAttr(self.name, *args, **kwargs)) or None

        if hasattr(meta, "hda"):
            return tuple(p.name() for p in meta.node(self.name).parms() or [])

        if hasattr(meta, "runtime"):
            return inspect.getmembers(meta.runtime.getnodebyname(self.name))

        if hasattr(meta, "data"):
            return tuple(inspect.getmembers(meta.data.objects[self.name]))

        if hasattr(meta, "knob"):
            return tuple([knob.name() for knob in meta.toNode(self.name).allKnobs()])

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.name).GetAttrs()

        if hasattr(meta, 'uclass'):
            return meta.assets.find_asset_data(self.name).get_asset()

        raise YException

    @trace
    def create(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            node_id = args[0] if "::" in args[0] else "::".join(["sbs", "compositing", args[0]])
            return YNode(meta.graph.newNode(node_id).getIdentifier())

        if hasattr(meta, "createNode"):
            return YNode(meta.createNode(*args, **kwargs))

        if hasattr(meta, "hda"):
            if len(args) == 0 and len(kwargs) == 0:
                return YNode(
                    partial(
                        meta.node(self.name).createNode, self.name
                    )(*args, **kwargs).path())
            return YNode(meta.node(self.name).createNode(*args, **kwargs).path())

        if hasattr(meta, "runtime"):
            obj = getattr(meta.runtime, args[0])
            msx_class = meta.runtime.classOf(obj)
            _obj = obj(**kwargs)

            if str(msx_class) == 'modifier':
                meta.runtime.addModifier(meta.runtime.getnodebyname(self.name), _obj)
                return YNode(meta.runtime.getnodebyname(self.name).name + "." + _obj.name)

            elif str(msx_class) == 'material':
                meta.runtime.material = _obj

            return YNode(_obj.name)

        if hasattr(meta, "data"):
            try:
                getattr(meta.ops.mesh, str(self).lower() + "_add")(*args, **kwargs)
            except AttributeError:
                getattr(meta.ops.object, str(self).lower() + "_add")(*args, **kwargs)

        if hasattr(meta, "fusion"):
            return YNode(meta.fusion.GetCurrentComp().AddTool(*args, **kwargs).Name)

        if hasattr(meta, 'uclass'):
            if not 'FactoryNew' in args[0]:
                return None
            # factory = getattr(meta, args[0])()
            # tool = meta.AssetToolsHelpers.get_asset_tools()
            # Name, Path, None
            # fargs = args[1:].append(factory)
            # tool.create_asset(*fargs)

        raise YException

    @trace
    def delete(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            return meta.graph.deleteNode(meta.graph.getNodeFromId(self.name))

        if hasattr(meta, "delete"):
            node = meta.toNode(self.name) if hasattr(meta, "knob") else self.name
            return meta.delete(node, *args, **kwargs)

        if hasattr(meta, "hda"):
            return meta.node(self.name).destroy()

        if hasattr(meta, "runtime"):
            return meta.runtime.delete(meta.runtime.getnodebyname(self.name))

        if hasattr(meta, "data"):
            return meta.context.scene.objects.unlink(meta.data.objects[self.name])

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.name).Delete()

        if hasattr(meta, 'uclass'):
            # return meta.assets.delete_asset(self.name)
            return meta.editor.get_actor_reference(self.name).destroy_actor()

        raise YException

    def instance(self, *args, **kwarg):
        if hasattr(meta, "SDNode"):
            return meta.graph.newInstanceNode(self.name, *args, **kwarg)

        if hasattr(meta, "instance"):
            if len(args) > 0:
                return meta.instance(self.name, lf=1)
            else:
                return meta.listRelatives(self.name, ap=1, f=1)[1:] or None

        if hasattr(meta, "runtime"):
            return YNode(meta.runtime.instance(meta.runtime.getnodebyname(self.name)).name)

        if hasattr(meta, "hda"):
            return meta.node(self.name).copyTo(*args, **kwarg)

        if hasattr(meta, "knob"):
            if len(args) > 0:
                return meta.clone(meta.toNode(self.name), *args, **kwarg)
            else:
                return meta.toNode(self.name).clones()

        if hasattr(meta, "fusion"):
            meta.fusion.GetCurrentComp().Copy(self.name)
            return meta.fusion.GetCurrentComp().Paste(*args, **kwarg)

        if hasattr(meta, 'uclass'):
            return

        raise YException

    def select(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            context = meta.sd_app.getLocationContext()
            if context:
                return context.getSelectedNodes()
            else:
                return None

        if hasattr(meta, "select"):
            if 'shape' not in kwargs and 's' not in kwargs:
                kwargs['s'] = True

            if len(args) == 0 and len(kwargs) == 0:
                return meta.ls(sl=1)
            else:
                return meta.select(*args, **kwargs)

        if hasattr(meta, "hda"):
            return meta.node(self.name).setCurrent(*args, **kwargs)

        if hasattr(meta, "runtime"):
            if len(args) == 0 and len(kwargs) == 0:
                return meta.runtime.select(meta.runtime.getnodebyname(self.name))
            else:
                if meta.runtime.execute('$') == meta.runtime.execute('$selection'):
                    return meta.runtime.execute('$ as array')
                else:
                    return meta.runtime.execute('$')

        if hasattr(meta, "knob"):
            if len(args) == 0 and len(kwargs) == 0:
                return meta.selectedNodes()
            else:
                return meta.toNode(self.name).setSelected()

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().CurrentFrame.FlowView.Select(*args, **kwargs)

        if hasattr(meta, 'uclass'):
            # return meta.editor.get_selected_assets()
            if len(args) == 0 and len(kwargs) == 0:
                return meta.editor.get_actor_reference(self.name, True)
            else:
                return meta.editor.get_selection_set()

        raise YException

    @trace
    def hide(self, on=True):
        if hasattr(meta, "data"):
            meta.data.objects[self.name].hide = on
            return

        if hasattr(meta, "runtime"):
            return getattr(meta.runtime, "hide" if on else "unhide")(meta.runtime.getnodebyname(self.name))

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.name).SetAttrs(
                {'TOOLB_Visible': on, "TOOLB_Locked": True})

        if hasattr(meta, 'uclass'):
            # set_actor_hidden_in_game
            return meta.editor.get_actor_reference(self.name).set_editor_property('hidden', on)

        raise YException

    @trace
    def geometry(self):
        if hasattr(meta, "ls"):
            dag = OM.MGlobal.getSelectionListByName(self.name).getDagPath(0)
            return OM.MFnMesh(dag)

        if hasattr(meta, "hda"):
            return meta.node(self.name).geometry()

        if hasattr(meta, "data"):
            return meta.data.meshes[self.name]

        if hasattr(meta, "runtime"):
            return meta.runtime.getnodebyname(self.name).mesh

        if hasattr(meta, 'uclass'):
            actor = meta.editor.get_actor_reference(self.name).get_class().get_name()
            if actor == 'StaticMeshActor' or actor == 'SkeletalMeshActor':
                return meta.editor.get_actor_reference(self.name)
            return

        raise YException


class YNode(YObject):
    """relationship object"""

    def __init__(self, item=None):
        if sys.version_info < (3, 2):
            super(YNode, self).__init__(item)
        else:
            super().__init__(item)
        self.item = item

        if self.item and hasattr(meta, "SDNode"):
            self._inputs = meta.graph.getNodeFromId(self.name).getProperties(meta.SDPropertyCategory.Input)
            self._outputs = meta.graph.getNodeFromId(self.name).getProperties(meta.SDPropertyCategory.Output)

    def parent(self, *args, **kwarg):
        if hasattr(meta, "SDNode"):
            nodes = []
            for prop in self._inputs:
                for connect in meta.graph.getNodeFromId(self.name).getPropertyConnections(prop):
                    nodes.append(YNode(connect.getInputPropertyNode().getIdentifier()))
            return nodes

        if hasattr(meta, "getAttr"):
            if len(args) == 0 and len(kwarg) > 0:
                return meta.parent(self.item, *args, **kwarg)
            else:
                return YNode(
                    partial(meta.listRelatives, self.item, p=1)(*args, **kwarg))

        if hasattr(meta, "hda"):
            return YNode(meta.node(self.name).parent().path())

        if hasattr(meta, "runtime"):
            if len(args) > 0:
                meta.runtime.getnodebyname(self.item).parent = args[0]
                return YNode(args[0])
            else:
                _parent = meta.runtime.getnodebyname(self.item).parent
                return YNode(_parent.name) if _parent else None

        if hasattr(meta, "knob"):
            index = meta.toNode(self.name).inputs() - 1
            return YNode(meta.toNode(self.name).input(index).name())

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.name).ParentTool

        if hasattr(meta, 'uclass'):
            return YNode(
                meta.editor.get_actor_reference(self.name).get_parent_actor().get_name())

        raise YException

    def children(self, *args, **kwarg):
        if hasattr(meta, "SDNode"):
            nodes = []
            for prop in self._outputs:
                for connect in meta.graph.getNodeFromId(self.name).getPropertyConnections(prop):
                    nodes.append(YNode(connect.getOutputPropertyNode().getIdentifier()))
            return nodes

        if hasattr(meta, "getAttr"):
            return partial(meta.listRelatives, self.item, c=1)(*args, **kwarg) or None

        if hasattr(meta, "hda"):
            return [YNode(node.name) for node in meta.node(self.item).children()]

        if hasattr(meta, "runtime"):
            if len(args) > 0:
                meta.runtime.execute('append $%s.children $%s' % (self.item, args[0]))
                return YNode(args[0])
            else:
                nodes = []
                children = meta.runtime.getnodebyname(self.item).children
                for i in range(children.count):
                    nodes.append(children[i].name)
                return nodes

        raise YException

    @trace
    def connect(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            args = (args[0], meta.graph.getNodeFromId(args[1].id), args[2])
            return meta.graph.getNodeFromId(self.name).newPropertyConnectionFromId(*args).getClassName()

        if hasattr(meta, "connectAttr"):
            return partial(meta.connectAttr, self.name + "." + args[0])(args[1:], **kwargs)

        if hasattr(meta, "hda"):
            return partial(meta.node(self.name).setInput, 0)(*args, **kwargs)

        if hasattr(meta, "knob"):
            return meta.toNode(self.name).setInput(*args, **kwargs)

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.name).ConnectInput(*args, **kwargs)

        raise YException

    @trace
    def disconnect(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            for arg in args:
                for prop in self._inputs:
                    if arg == prop.getId():
                        return meta.graph.getNodeFromId(self.name).deletePropertyConnections(prop)
                for prop in self._outputs:
                    if arg == prop.getId():
                        return meta.graph.getNodeFromId(self.name).deletePropertyConnections(prop)
            return

        if hasattr(meta, "disconnectAttr"):
            return partial(meta.disconnectAttr, self.name + "." + args[0])(args[1:], **kwargs)

        if hasattr(meta, "hda"):
            return partial(meta.node(self.name).setInput, 0, None)(*args, **kwargs)

        if hasattr(meta, "knob"):
            return meta.toNode(self.name).setInput(0, None)

        if hasattr(meta, "fusion"):
            return setattr(meta.fusion.GetCurrentComp().FindTool(self.name), "Input", None)

        raise YException

    @trace
    def inputs(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            return [connect.getId() for connect in self._inputs if connect.isConnectable()]

        if hasattr(meta, "listConnections"):
            return partial(meta.listConnections, s=1)(*args, **kwargs)

        if hasattr(meta, "hda"):
            return [YNode(node.name) for node in meta.node(self.name).inputs()]

        if hasattr(meta, "knob"):
            return [YNode(meta.toNode(self.name).input(index).name()) for index in
                    range(meta.toNode(self.name).inputs())]

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.name).GetInputList().values()[0].GetAttrs()

        raise YException

    @trace
    def outputs(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            return [connect.getId() for connect in self._outputs if connect.isConnectable()]

        if hasattr(meta, "listConnections"):
            return partial(meta.listConnections, d=1)(*args, **kwargs)

        if hasattr(meta, "hda"):
            return [YNode(node.name) for node in meta.node(self.name).outputs()]

        if hasattr(meta, "knob"):
            return meta.toNode(self.name).dependencies(meta.EXPRESSIONS)

        if hasattr(meta, "fusion"):
            return meta.fusion.GetCurrentComp().FindTool(self.name).GetOutputList().values()[0].GetAttrs()

        raise YException


@total_ordering
class YAttr(_YAttr):
    """parametric object"""

    def __init__(self, *args):
        assert len(args) > 2, "parameter is invalid."
        self._values = args
        self.obj, self.val = self._values[1:]

    @property
    def value(self):
        if hasattr(meta, "SDNode"):
            return self._values[0]
        
        if ':' in str(self._values[0]):
            return getattr(self._values[0], self.val)
        else:
            return self._values[0]

    def __getitem__(self, idx):
        return self._values[idx]

    def __repr__(self):
        return str(self.value)

    def __eq__(self, other):
        return self.value == other.value

    def __gt__(self, other):
        return self.value >= other.value

    @trace
    def set(self, *args, **kwargs):
        if hasattr(meta, "SDNode"):
            sd_value = meta.SDValueInt.sNew(args[0])

            if type(self.value) == bool:
                sd_value = meta.SDValueBool.sNew(args[0])
            if type(self.value) == float:
                sd_value = meta.SDValueFloat.sNew(args[0])
            if type(self.value) == str:
                sd_value = meta.SDValueString.sNew(args[0])

            prop = meta.graph.getNodeFromId(self.obj).getPropertyFromId(self.val, meta.SDPropertyCategory.Input)
            return meta.graph.getNodeFromId(self.obj).setPropertyValue(prop, sd_value)

        if hasattr(meta, "setAttr"):
            return meta.setAttr(self.obj + "." + self.val, *args, **kwargs)

        if hasattr(meta, "hda"):
            parm = (meta.node(self.obj).parm(self.val) or
                    meta.node(self.obj).parmTuple(self.val))
            return parm.set(
                args[0].tolist() if hasattr(args[0], "T") else args[0], **kwargs)

        if hasattr(meta, "runtime"):
            if ':' in str(self._values[0]):
                return setattr(self._values[0], self.val, args[0])
            else:
                return setattr(meta.runtime.getnodebyname(self.obj), self.val, args[0])

        if hasattr(meta, "data"):
            return setattr(meta.data.objects[self.obj],
                           self.val, args[0].tolist() if hasattr(args[0], "T") else args[0])

        if hasattr(meta, "knob"):
            return meta.toNode(self.obj)[self.val].setValue(args[0], **kwargs)

        if hasattr(meta, "fusion"):
            return setattr(meta.fusion.GetCurrentComp().FindTool(self.obj), self.val, args[0])

        if hasattr(meta, 'uclass'):
            return meta.assets.find_asset_data(self.obj).get_asset().set_editor_property(self.obj, self.val)

        raise YException

    def create(self, *args, **kwargs):
        pass

    def delete(self, *args, **kwargs):
        pass

    @trace
    def lock(self, on):
        if hasattr(meta, "setAttr"):
            return meta.setAttr(self.obj + "." + self.val, lock=on)

        if hasattr(meta, "hda"):
            return meta.node(self.obj).parm(self.val).lock(on)

        if hasattr(meta, "knob"):
            return meta.toNode(self.obj)[self.val].setEnabled(on)

        raise YException

    @trace
    def hide(self, on=True):
        if hasattr(meta, "setAttr"):
            return meta.setAttr(self.obj + "." + self.val, keyable=not on, channelBox=not on)

        if hasattr(meta, "hda"):
            return meta.node(self.obj).parm(self.val).hide(on)

        if hasattr(meta, "knob"):
            return meta.toNode(self.obj)[self.val].setVisible(not on)

        raise YException

    @property
    def asVector(self):
        try:
            return YVector(self._values[0])
        except TypeError:
            return YVector(*self._values[0])

    @property
    def asColor(self):
        try:
            return YColor(self._values[0])
        except TypeError:
            return YColor(*self._values[0])

    @property
    def asMatrix(self):
        try:
            return YMatrix(self._values[0])
        except TypeError:
            return YMatrix(*self._values[0])


class YFile(_YObject):
    """save, open and export"""

    def __init__(self, path=""):
        self.file = path

    @property
    def name(self):
        return os.path.basename(self.file)

    @property
    def path(self):
        return os.path.dirname(self.file)

    @classmethod
    def open(cls, *args, **kwargs):
        from yurlungur.core.command import file

        if args[0].endswith("abc"):
            return cls(file.abcImporter(*args, **kwargs))

        if args[0].endswith("fbx"):
            return cls(file.fbxImporter(*args, **kwargs))

        if hasattr(meta, "sbs"):
            return cls(meta.manager.loadUserPackage(*args, **kwargs))

        if hasattr(meta, "file"):
            return cls(partial(meta.file, i=1)(*args, **kwargs))

        if hasattr(meta, "hipFile"):
            return cls(meta.hipFile.load(*args, **kwargs))

        if hasattr(meta, "runtime"):
            if meta.runtime.loadMaxFile(*args, **kwargs):
                return cls(args[0])

        if hasattr(meta, "data"):
            return cls(meta.ops.wm.open_mainfile(*args, **kwargs))

        if hasattr(meta, "knob"):
            return meta.scriptOpen(*args, **kwargs)

        if hasattr(meta, "fusion"):
            storage = meta.resolve.GetMediaStorage()
            if "." in args:
                return storage.AddItemsToMediaPool(*args)
            else:
                return meta.manager.LoadProject(*args, **kwargs)

        raise YException

    @classmethod
    def save(cls, *args, **kwargs):
        from yurlungur.core.command import file

        if args[0].endswith("abc"):
            return cls(file.abcExporter(*args, **kwargs))

        if args[0].endswith("fbx"):
            return cls(file.fbxExporter(*args, **kwargs))

        if hasattr(meta, "sbs"):
            return cls(meta.manager.savePackageAs(*args, **kwargs))

        if hasattr(meta, "file"):
            return cls(partial(meta.file, s=1)(*args, **kwargs))

        if hasattr(meta, "hipFile"):
            return cls(meta.hipFile.save(*args, **kwargs))

        if hasattr(meta, "runtime"):
            if meta.runtime.saveMaxFile(*args, **kwargs):
                return cls(args[0])

        if hasattr(meta, "data"):
            return meta.ops.wm.save_mainfile(*args, **kwargs)

        if hasattr(meta, "knob"):
            return meta.scriptSave(*args, **kwargs)

        if hasattr(meta, "fusion"):
            return meta.manager.SaveProject(*args, **kwargs)

        raise YException

    @property
    def current(self):
        if hasattr(meta, "sbs"):
            return meta.manager.getUserPackageFromFilePath()

        if hasattr(meta, "file"):
            return meta.file(exn=1, q=1)

        if hasattr(meta, "hipFile"):
            return meta.hipFile.path()

        if hasattr(meta, "runtime"):
            return meta.runtime.maxFilePath + meta.runtime.maxFileName

        if hasattr(meta, "data"):
            return meta.data.filepath

        if hasattr(meta, "knob"):
            return meta.scriptName()

        if hasattr(meta, "fusion"):
            return meta.manager.GetCurrentProject()

        raise YException

    def reference(self):
        raise YException



# -*- coding: utf-8 -*-
from functools import partial

from yurlungur.core.proxy import YNode, YFile
from yurlungur.tool.meta import meta
from yurlungur.core.wrapper import YException

__all__ = ["file", "cmd"]


class Command(object):
    @staticmethod
    def register(func):
        pass

    @staticmethod
    def unregister(func):
        pass

    @classmethod
    def list(cls):
        return [obj for obj in dir(cls) if not obj.startswith("_")]


def _ls(cls, *args, **kwargs):
    gen = meta.ls(*args, **kwargs) if hasattr(meta, "ls") else meta.pwd().allItems()
    return tuple(YNode(obj) for obj in gen)


def _rm(cls, *args):
    for obj in args:
        YNode(obj).delete()


def _glob(cls, *args, **kwargs):
    gen = meta.ls(*args, **kwargs) if hasattr(meta, "ls") else meta.pwd().glob(*args, **kwargs)
    return tuple(YNode(obj) for obj in gen)


def _select(cls, *args, **kwargs):
    if hasattr(meta, "select"):
        meta.select(*args, **kwargs)

    if hasattr(meta, "hda"):
        for node in meta.nodes(args):
            node.setSelected(True, **kwargs)


def _alembicImporter(cls, *args, **kwargs):
    """
    >>> f = YFile()
    >>> YFile.new_method = new_method
    >>> print f.new_method("new")
    """
    if hasattr(meta, "AbcImport"):
        return cls(meta.AbcImport(*args, **kwargs))

    if hasattr(meta, "hda"):
        geo = yr.YNode("obj").create("geo")
        abc = geo.create("alembic")
        abc.fileName.set(*args)
        return cls(*args)

    if hasattr(meta, "runtime"):
        importer = partial(meta.runtime.importFile, args[0], meta.runtime.Name("noPrompt"), using='AlembicImport')
        if importer(**kwargs):
            return args[0]

    raise YException


def _alembicExporter(cls, *args, **kwargs):
    if hasattr(meta, "AbcExport"):
        return cls(meta.AbcExport(*args, **kwargs))

    if hasattr(meta, "runtime"):
        export = partial(meta.runtime.exportFile, args[0], meta.runtime.Name("noPrompt"), using='AlembicExport')
        if export(**kwargs):
            return args[0]

    raise YException


def _fbxImporter(cls, *args, **kwargs):
    if hasattr(meta, "importFBX"):
        return meta.importFBX(*args, **kwargs)

    if hasattr(meta, "runtime"):
        importer = partial(meta.runtime.importFile, args[0], meta.runtime.Name("noPrompt"), using='FBXIMPORTER')
        if importer(**kwargs):
            return args[0]

    if hasattr(meta, 'AssetImportTask'):
        meta.AssetToolsHelpers.get_asset_tools().import_asset_task()

    if hasattr(meta, 'eval'):
        return cls(meta.eval("FBXImport -file {0};".format(*args)))

    raise YException


def _fbxExporter(cls, *args, **kwargs):
    if hasattr(meta, "runtime"):
        export = partial(meta.runtime.exportFile, args[0], meta.runtime.Name("noPrompt"), using='FBXEXPORTER')
        if export(**kwargs):
            return args[0]

    if hasattr(meta, 'eval'):
        return cls(meta.eval("FBXExportInAscii -v true; FBXExport -f \"{}\" -s;".format(*args)))

    raise YException


# Monkey-Patch
file = YFile()
YFile.abcImporter = _alembicImporter
YFile.abcExporter = _alembicExporter
YFile.fbxImporter = _fbxImporter
YFile.fbxExporter = _fbxExporter

cmd = Command()
Command.ls = _ls
Command.rm = _rm
Command.glob = _glob
Command.select = _select



# -*- coding: utf-8 -*-
from yurlungur.core.wrapper import YMObject
from yurlungur.tool import util

meta = YMObject()
modules = util.__make_completer(meta.module)

__all__ = ["meta"]

# add completer
# for mod in modules:
#     __all__.append(mod)



import doctest
import unittest

import sys

sys.path.append('../yurlungur')

import yurlungur as yr
from yurlungur.core.env import installed, Maya

@unittest.skipUnless(installed("maya"), "Maya is not found")
class TestMaya(unittest.TestCase):
    def test_env(self):
        assert Maya()

    def test_maya(self):
        standalone.mayapy("import yurlungur as yr; print yr.name")

    @unittest.skip("only runtime")
    def test_cmds(self):
        yr.meta.file(new=1, f=1)
        cone, shape = yr.meta.polyCone()
        node = yr.YObject(shape)
        node.attr("castsShadows").set(1)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import sys

sys.path.append('../yurlungur')

import doctest
import unittest
from yurlungur.core import app


class TestCore(unittest.TestSuite):

    def test_app(self):
        self.assertTrue(app.application == "standalone")

    def test_vector(self):
        self.assertTrue(yr.YVector() == yr.YVector())


if __name__ == '__main__':
    unittest.main()



# coding: utf-8
import unreal


editor = EditorUtil()
assets = GetEditorAssetLibrary()
materials = MaterialEditingLib()
anims = GetAnimationLibrary()


def apply_lods(static_mesh):
    number_of_vertices = unreal.EditorStaticMeshLibrary.get_number_verts(static_mesh, 0)
    if number_of_vertices < 10:
        return

    print("treating asset: " + static_mesh.get_name())
    print("existing LOD count: " + str(unreal.EditorStaticMeshLibrary.get_lod_count(static_mesh)))

    options = unreal.EditorScriptingMeshReductionOptions()
    options.reduction_settings = [unreal.EditorScriptingMeshReductionSettings(1.0, 1.0),
                                  unreal.EditorScriptingMeshReductionSettings(0.8, 0.75),
                                  unreal.EditorScriptingMeshReductionSettings(0.6, 0.5),
                                  unreal.EditorScriptingMeshReductionSettings(0.4, 0.25)
                                  ]
    options.auto_compute_lod_screen_size = False
    unreal.EditorStaticMeshLibrary.set_lods(static_mesh, options)
    unreal.EditorAssetLibrary.save_loaded_asset(static_mesh)
    print("new LOD count: " + str(unreal.EditorStaticMeshLibrary.get_lod_count(static_mesh)))


all_assets = unreal.EditorAssetLibrary.list_assets(asset_path)
all_assets_loaded = [unreal.EditorAssetLibrary.load_asset(a) for a in all_assets]
static_mesh_assets = unreal.EditorFilterLibrary.by_class(all_assets_loaded, unreal.StaticMesh)
map(apply_lods, static_mesh_assets)

allAssets = assets.list_assets("/Game/", True, False)
for asset in allAssets or []:
    deps = assets.find_package_referencers_for_asset(asset, False)
    if (len(deps) <= 0):
        assets.delete_asset(asset)

selectedAssets = editor.get_selected_assets()

selectedAssetName = selectedAssets.get_name()
selectedAssetPath = selectedAssets.get_path_name()
selectedAssetClass = selectedAssets.get_class()
allAssets = editor.list_assets(workingPath, True, False)

for selectedAsset in selectedAssets:
    selectedAsset.modify(True)
    anims.remove_all_animation_notify_tracks(selectedAsset)

editor.load_asset()

import sys

sys.path.append("C:/Users/sho/Downloads/Yurlungur")
import yurlungur as yr

reload(yr)
print yr.YNode('/Game/aaaa').name

if hasattr(meta, 'uclass'):
    meta.editor.get_actor_reference(self.name)

@unreal.uclass()
class EditorUtil(unreal.GlobalEditorUtilityBase):
    pass


@unreal.uclass()
class GetEditorAssetLibrary(unreal.EditorAssetLibrary):
    pass


@unreal.uclass()
class MaterialEditingLib(unreal.MaterialEditingLibrary):
    pass


@unreal.uclass()
class GetAnimationLibrary(unreal.AnimationLibrary):
    pass


class Project(object):
    def __init__(self, project):
        self.project = project

    @property
    def sequences(self):
        return Timeline(self.project)


class Timeline(object):
    def __init__(self, timeline):
        self.timeline = timeline

    @property
    def tracks(self):
        return Track(self.timeline)


class Track(object):
    def __init__(self, track):
        self.track = track

    @property
    def clips(self):
        return Item(self.track)


class Item(object):
    pass



# -*- coding: utf-8 -*-
import platform
import sys
from yurlungur.core.env import __import__

application = sys.executable
__doc__ = "app modules"


def exApplication(module=""):
    application = ""

    if __import__(module):
        application = __import__(module)

    elif application == "":
        from yurlungur.tool import standalone

        application = standalone

    return application


if "maya" in application:
    import maya.cmds as cmds

    application = cmds

elif __import__("hou"):
    import hou

    application = hou

elif "Substance" in application:
    import sd.api as sdapi

    application = sdapi


elif "UE4Editor" in application:
    import unreal

    application = unreal

elif "3dsmax" in application:
    import pymxs

    application = pymxs

elif "blender" in application:
    import bpy

    application = bpy

elif "UE4Editor" in application:
    import unreal

    application = unreal

elif "Nuke" in application:
    import nuke

    application = nuke

elif __import__("DaVinciResolveScript"):
    import DaVinciResolveScript

    application = DaVinciResolveScript

else:
    if platform.python_implementation() == 'IronPython':
        import clr

        clr.AddReferenceByPartialName('UnityEngine')
        import UnityEngine as unity

        application = unity
    else:
        application = exApplication()

__all__ = ["application", "exApplication"]



import doctest
import unittest

import sys

sys.path.append('../yurlungur')

import yurlungur as yr
from yurlungur.core.env import installed, Substance


@unittest.skipUnless(installed("substance"), "Substance is not found")
class TestMaya(unittest.TestCase):
    def test_env(self):
        assert Substance()

    def test_maya(self):
        standalone.mayapy("import yurlungur as yr; print yr.name")

    @unittest.skip("only runtime")
    def test_cmds(self):
        uniform = yr.YNode().create("uniform")

        node = yr.YNode("Directional Warp")
        print(node, node.attr("$pixelsize"), node.attrs)
        normal = yr.YNode("Normal")
        print(normal)

        node.connect("unique_filter_output", normal, "inputNodeOutput.connector")
        print(node.parent())
        node.disconnect("unique_filter_output")
        print(node.parent())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import functools
import inspect
import os
import sqlite3
import sys
import time
import traceback

import yurlungur
from yurlungur.core import logger


def cache(func, *args, **kwargs):
    saved = {}

    @functools.wraps(func)
    def wrapper(*args):
        if args in saved:
            return saved[args]
        result = func(*args)
        saved[args] = result
        return result

    return wrapper if sys.version_info < (3, 2) else functools.lcu_cache(*args, **kwargs)


def trace(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            if hasattr(logger.logger, "warn"):
                logger.logger.warn(traceback.format_exc())
            else:
                logger.logger.log(traceback.format_exc(), logger.Warning)

    return wrapper


def timer(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        yurlungur.logger.log(
            '{0} start'.format(func.__name__)
        )
        start_time = time.clock()
        ret = func(*args, **kwargs)
        end_time = time.clock()
        yurlungur.logger.log(
            '\n{0}: {1:,f}s'.format("total: ", (end_time - start_time))
        )
        return ret

    return wrapper


def __db_loader():
    from collections import namedtuple

    local = os.path.dirname(os.path.dirname(inspect.currentframe().f_code.co_filename))
    cache = os.path.join(local, "user", "cache.db").replace(os.sep, "/")

    conn = sqlite3.connect(cache)
    c = conn.cursor()
    for row in c.execute('SELECT * FROM stocks ORDER BY price'):
        print(row)

    return namedtuple


def __db_attr():
    local = os.path.dirname(os.path.dirname(inspect.currentframe().f_code.co_filename))
    cache = os.path.join(local, "user", "cache.db").replace(os.sep, "/")

    conn = sqlite3.connect(cache)
    c = conn.cursor()
    c.execute('''CREATE TABLE stocks
                 (date text, trans text, symbol text, qty real, price real)'''
              )
    c.execute(
        "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14)"
    )
    conn.commit()
    conn.close()


def __make_completer(mod):
    local = os.path.dirname(os.path.dirname(inspect.currentframe().f_code.co_filename))
    completer = os.path.join(local, "user", "completer.pyi").replace(os.sep, "/")

    header = "\"\"\"this document generated by internal module.\"\"\"\n\n\n"
    module = __import__(mod)
    with open(completer, "w") as f:
        f.write(header)

        for fn, _ in inspect.getmembers(module):
            if fn.startswith("_"):
                continue

            f.write("def {0}(*args, **kwargsargs):\n".format(fn))
            f.write("   \"\"\"{0}\"\"\"\n".format(inspect.getdoc(fn)))
            f.write("   pass\n\n")

    # return inspect.getmembers(__import__(completer))



# coding: utf-8
import sys
import unittest

sys.path.append('../yurlungur')
import yurlungur as yr
from yurlungur.adapters import resolve


class TestResolve(unittest.TestSuite):

    def test_aaa(self):
        projects = resolve.Projects()
        print projects.timelines.tracks.comps
        print projects["Hellow"].timelines["g"].tracks

    def test_fscript(self):
        # comp起点とfusion起点の２つ
        # https://www.steakunderwater.com/VFXPedia/96.0.243.189/index1bea.html?title=Eyeon:Script/Reference/Applications/Fusion/Classes/Composition
        # print comp.GetToolList().values()[0].GetAttrs()
        # fpath = fusion.MapPath("/Users/shosumioka/Documents/DaVinci Resolve/aaa.comp")
        # mcomp = fusion.LoadComp(fpath)
        # pprint.pprint(mcomp.GetAttrs()['COMPS_FileName'])
        # mcomp.Save(), mcomp.Print(), mcomp.Close(), mcomp.Lock(), mcomp.Unlock(), mcomp.Paste()
        comp = resolve.Fusion().GetCurrentComp()
        bg1 = comp.Background()
        pl1 = comp.Plasma()
        mg1 = comp.Merge({"Background": bg1, "Foreground": pl1})
        comp.FindTool("MediaOut1").Input.ConnectTo(mg1.Output)
        for n in comp.ActiveTool.GetChildrenList().values():
            print n.Name
        # print comp.FindTool("MediaOut1").GetInputList()
        # comp.FindTool("MediaOut1").Input.ConnectTo()
        # print yr.YNode(mg1.Name).parent(), yr.YNode(mg1.Name).children()
        # yr.YNode("Loader1").connect()

    def test_fusion(self):
        mg1 = yr.YNode().create("Merge")
        print yr.YNode(mg1).children()
        paint = yr.YNode().create("Paint", 2, 1)
        print paint.id, paint.name, paint.attrs
        paint.hide()
        paint.instance()
        paint.select()
        paint.connect(yr.YNode("MediaOut1"))
        yr.YNode().create("Grain").delete()
        myblur = yr.YNode().create("Blur")
        mybg = yr.YNode().create("Background")
        mybg.connect(yr.YNode(myblur.name))
        mybg.Type.set("Vertical")
        myblur.XBlurSize.set(5.0)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import contextlib
import functools

from yurlungur.tool.meta import meta
from yurlungur.core import env


class UndoGroup(object):
    def __init__(self, label):
        self.label = label

    def __enter__(self):
        if env.Maya():
            meta.undoInfo(ock=1)
            return self
        elif env.Blender():
            self.undo = meta.context.user_preferences.edit.use_global_undo
            meta.context.user_preferences.edit.use_global_undo = False
        elif env.Davinci():
            meta.fusion.StartUndo()

    def __exit__(self, exc_type, exc_value, traceback):
        if env.Maya():
            meta.undoInfo(cck=1)
        elif env.Blender():
            meta.context.user_preferences.edit.use_global_undo = self.undo
        elif env.Davinci():
            meta.fusion.EndUndo()


if env.Houdini():
    UndoGroup = meta.undos.group

if env.Unreal():
    UndoGroup = meta.ScopedEditorTransaction

if env.Max():
    UndoGroup = functools.partial(meta.undo, True)

if env.Nuke():
    UndoGroup = meta.Undo



import doctest
import unittest
import sys

sys.path.append('../yurlungur')

from yurlungur.core.env import installed, Unreal


@unittest.skipUnless(installed("unreal"), "Unreal is not found")
class TestUnreal(unittest.TestCase):
    def test_app(self):
        return True

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import sys
import os
import inspect
import fnmatch
import traceback

import yurlungur
from yurlungur.core import env

if env.Qt():
    from yurlungur.Qt.QtCore import *
    from yurlungur.Qt.QtGui import *
    from yurlungur.Qt.QtWidgets import *


class OpenGL(object):
    """openGL wrapper"""

    def __getattr__(self, item):
        def _getGL(mod):
            for cmd, _ in inspect.getmembers(mod):
                if fnmatch.fnmatch(item, "".join(["*", cmd])):
                    setattr(
                        self, cmd,
                        (lambda str: dict(inspect.getmembers(mod))[str])(cmd)
                    )
                    return getattr(self, item)

        _tmp = []

        if env.Maya():
            from maya import OpenMayaRender as _mgl
            _tmp.extend(_mgl, _mgl.MHardwareRenderer.theRenderer().glFunctionTable())

        if env.Blender():
            import bgl
            _tmp.extend(bgl)

        for gl in _tmp:
            if _getGL(gl):
                return _getGL(gl)

        try:
            from OpenGL import GL as gl
            return gl
        except ImportError:
            import ctypes
            return ctypes.cdll.OpenGL32


@env.Qt
def widgetPtr():
    import yurlungur.core.app
    app_name = yurlungur.core.application.__name__

    if app_name == "maya.cmds":
        from yurlungur.Qt import QtCompat
        from maya import OpenMayaUI
        ptr = long(OpenMayaUI.MQtUtil.mainWindow())
        return QtCompat.wrapInstance(ptr, QWidget)

    if app_name == "pymxs":
        import MaxPlus
        return MaxPlus.QtHelpers_GetQmaxMainWindow()

    if app_name == "hou":
        import hou
        return hou.qt.mainWindow()

    if app_name == "nuke":
        return QApplication.activeWindow()

    return None


@env.Qt
def show(view):
    try:
        view.deleteLater()
    except:
        yurlungur.logger.log(view)

    try:
        __dark_view(view)

        if env.Max() or env.Unreal():
            __max_protect_show(view)
        else:
            view.show()
    except:
        view.deleteLater()
        yurlungur.logger.warn(traceback.print_exc())

    if not QApplication.instance():
        app = QApplication(sys.argv)
        __dark_view(view)

        if env.Max():
            __max_protect_show(view)
        else:
            view.show()

        sys.exit(app.exec_())


class __GCProtector(object):
    widgets = []


def __max_protect_show(w):
    w.setWindowFlags(Qt.WindowStaysOnTopHint)
    w.show()
    __GCProtector.widgets.append(w)


def __dark_view(view):
    local = os.path.dirname(
        os.path.dirname(inspect.currentframe().f_code.co_filename)
    )
    with open(local + "/user/dark.css") as f:
        view.setStyleSheet("".join(f.readlines()))



# coding: utf-8
import sys
import unittest

sys.path.append('../yurlungur')
import yurlungur as yr


class TestResolve(unittest.TestSuite):

    def test_node(self):
        yr.YNode("").create("TimeWarp")

        if hasattr(yr.meta, "knob"):
            import nuke
            b = nuke.toNode("aaa")
            k = nuke.Array_Knob("name", "label")
            b.addKnob(k)



# -*- coding: utf-8 -*-
from __future__ import print_function
import sys
from logging import (
    getLogger, Handler, INFO, WARNING, basicConfig
)
from pprint import pformat

import yurlungur
from yurlungur.core import env


class GuiLogHandler(Handler):
    def __init__(self, *args, **kwargs):
        super(GuiLogHandler, self).__init__(*args, **kwargs)

        if env.Maya():
            import maya.OpenMaya as om
            self.MGlobal = om.MGlobal

    def emit(self, record):
        from yurlungur.tool.meta import meta

        msg = self.format(record)
        if record.levelno > WARNING:
            if env.Maya():
                self.MGlobal.displayError(msg)
            elif env.Houdini():
                meta.ui.setStatusMessage(msg, severity=meta.severityType.Error)
            elif env.Unreal():
                meta.log_error(msg)
            elif env.Unity():
                meta.Debug.LogError(msg)
            elif env.Nuke():
                meta.error(msg)
            elif env.Max():
                meta.print_(msg, True, False)

        elif record.levelno > INFO:
            if env.Maya():
                self.MGlobal.displayWarning(msg)
            elif env.Houdini():
                meta.ui.setStatusMessage(msg, severity=meta.severityType.Warning)
            elif env.Unreal():
                meta.log_warning(msg)
            elif env.Unity():
                meta.Debug.LogWarning(msg)
            elif env.Nuke():
                meta.warning(msg)
            elif env.Max():
                meta.print_(msg, True, False)

        else:
            if env.Maya():
                self.MGlobal.displayInfo(msg)
            elif env.Houdini():
                meta.ui.setStatusMessage(msg, severity=meta.severityType.Message)
            elif env.Unreal():
                meta.log(msg)
            elif env.Unity():
                meta.Debug.Log(msg)
            elif env.Nuke():
                meta.debug(msg)
            elif env.Max():
                meta.print_(msg, False, True)


if env.Substance():
    import sd
    import sd.logger as slog

    logger = sd.getContext().getLogger()
    Warning = slog.LogLevel.Warning

elif env.Blender():
    pass  # TODO: https://code.blender.org/2016/05/logging-from-python-code-in-blender/

else:
    logger = getLogger(yurlungur.__name__)
    logger.setLevel(INFO)
    handler = GuiLogHandler()
    logger.addHandler(handler)
    basicConfig(level=INFO, stream=sys.stdout)


def log(*msgs):
    logger.log(pformat(*msgs))



# coding: utf-8
import bpy


class Project(object):
    def __init__(self, project):
        self.project = project

    @property
    def sequences(self):
        return Timeline(self.project)


class Timeline(object):
    def __init__(self, timeline):
        self.timeline = timeline

    @property
    def tracks(self):
        return Track(self.timeline)


class Track(object):
    def __init__(self, track):
        self.track = track

    @property
    def clips(self):
        return Item(self.track)


class Item(object):
    pass



# encoding: utf-8
from __future__ import print_function
import sys
import os

PORT = 10000  # sample
USER = os.environ.get("APPDATA")
PTH = [p for p in os.listdir(USER) if os.path.isdir("{}/{}".format(USER, p))]
EGG = "{}/{}/debug-eggs/pycharm-debug.egg".format(USER, PTH[0 if len(PTH) == 1 else -1])
PDV = "{}/{}/helpers/pydev".format(USER, PTH[0 if len(PTH) == 1 else -1])

for lib in [EGG, PDV]:
    if lib not in sys.path:
        sys.path.insert(0, lib)

import pydevd


def main():
    try:
        try:
            import maya.cmds as cmds
            import maya.utils

            if not cmds.commandPort(':{}'.format(PORT), q=1):
                cmds.commandPort(n=':{}'.format(PORT))
                # maya.utils.executeDeferred(cmds.commandPort(n=':{}'.format(PORT)))
            print(cmds.commandPort(':{}'.format(PORT), q=1))

        except ImportError:
            import hou, hrpyc
            hrpyc.start_server(PORT, False, False)

        pydevd.stoptrace()
        pydevd.settrace(
            "localhost", port=PORT - 1, stdoutToServer=True, stderrToServer=True
        )
        print("trace start")

    except:
        pydevd.stoptrace()
        cmds.commandPort(n=':{}'.format(PORT), close=1)
        print("trace stop", "maya port {}".format(cmds.commandPort(':{}'.format(PORT), q=1)))


if __name__ == '__main__':
    main()



from __future__ import print_function, unicode_literals, absolute_import

__all__ = []
__version__ = "0.9.5"

import sys

assert sys.version_info > (2, 7), ('yurlungur currently requires Python 2.7 later')
sys.dont_write_bytecode = True

from yurlungur.core import * # noQA
from yurlungur.tool.math import (
    YVector, YColor, YMatrix
)
from yurlungur.tool.util import (
    cache, trace, timer
)
from yurlungur.tool import ui
from yurlungur.tool.meta import meta
from yurlungur.adapters import *  # noQA

# info
name = __name__
version = __version__


import doctest
import unittest

import sys

sys.path.append('../yurlungur')

from yurlungur.core.env import installed, Houdini

@unittest.skipUnless(installed("houdini"), "Houdini is not found")
class TestHoudini(unittest.TestCase):
    def test_env(self):
        self.assertTrue(installed("houdini"))

    def test_hou(self):
        standalone.hython("import yurlungur as yr; print yr.name, yr.version")

    def test_node(self):
        standalone.hython("import yurlungur as yr; print yr.YNode('obj')")

    def test_attr(self):
        standalone.hython("import yurlungur as yr; yr.YNode('obj').create('geo').tx.set(5)")

    def test_file(self):
        standalone.hython("import yurlungur as yr; yr.YFile.save('temp.hip')")


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import inspect

import yurlungur
from yurlungur.core import app, env


class YException(NotImplementedError):
    """
    >>> raise NotImplementedError(app.application)
    """
    pass


class YMObject(object):
    """command wrapper for any application"""

    if env.Substance():
        import sd
        from sd.api.sdproperty import SDPropertyCategory
        context = sd.getContext()
        sd_app = context.getSDApplication()
        manager = sd_app.getPackageMgr()
        graph = manager.getUserPackages()[0].getChildrenResources(True)[0]

    if env.Davinci():
        resolve = env.__import__("DaVinciResolveScript").scriptapp("Resolve")
        if resolve:
            manager = resolve.GetProjectManager()
            fusion = resolve.Fusion()

    if env.Unreal():
        from yurlungur.adapters import unreal as ue
        editor = ue.EditorUtil()
        assets = ue.GetEditorAssetLibrary()
        materials = ue.MaterialEditingLib()
        anim = ue.GetAnimationLibrary()

    def __getattr__(self, item):
        for cmd, _ in inspect.getmembers(app.application):
            if cmd == item:
                setattr(
                    yurlungur, cmd,
                    (lambda str: dict(inspect.getmembers(app.application))[str])(cmd)
                )
                return getattr(yurlungur, item)

        raise YException

    def eval(self, script):
        if env.Maya():
            import maya.mel as mel
            return mel.eval(script)
        if env.Houdini():
            return app.application.hscript(script)
        if env.Max():
            import MaxPlus
            return MaxPlus.Core.EvalMAXScript(script)
        if env.Davinci() and self.resolve:
            self.fusion.GetCurrentComp().Execute(script)

        raise YException

    @property
    def module(self):
        return app.application.__name__


class MetaObject(type):
    def __new__(cls, name, bases, attrs):
        return super(MetaObject, cls).__new__(cls, name, bases, attrs)


class MetaAttr(type):
    def __new__(cls, name, bases, attrs):
        return super(MetaAttr, cls).__new__(cls, name, bases, attrs)


class MetaNode(type):
    def __new__(cls, name, bases, attrs):
        return super(MetaNode, cls).__new__(cls, name, bases, attrs)


# Dynamic Class
_YObject = MetaObject("YObject", (object,), {"__doc__": MetaObject.__doc__})
_YNode = MetaNode("YNode", (object,), {"__doc__": MetaNode.__doc__})
_YAttr = MetaAttr("YAttr", (object,), {"__doc__": MetaAttr.__doc__})
_YVector = _YMatrix = _YColor = OM = object

if env.Maya():
    import maya.api.OpenMaya as OM

    _YVector = type('_YVector', (OM.MVector,), dict())
    _YMatrix = type('_YMatrix', (OM.MMatrix,), dict())
    _YColor = type('_YColor', (OM.MColor,), dict())

    meta = YMObject()
    if hasattr(meta, "loadPlugin"):
        for plugin in "fbxmaya.mll", "AbcImport.mll", "AbcExport.mll":
            meta.loadPlugin(plugin, qt=1)

elif env.Houdini() or env.Unreal():
    meta = YMObject()

    _YVector = type('_YVector', (
        meta.Vector if hasattr(meta, "Vector") else meta.Vector3,
    ), dict())

    _YMatrix = type('_YMatrix', (
        meta.Matrix if hasattr(meta, "Matrix") else meta.Matrix4,
    ), dict())

    _YColor = type('_YColor', (meta.Color,), dict())

elif env.Blender():
    import mathutils

    _YVector = type('_YVector', (mathutils.Vector,), dict())
    _YMatrix = type('_YMatrix', (mathutils.Matrix,), dict())
    _YColor = type('_YColor', (mathutils.Color,), dict())

elif env.Substance():
    meta = YMObject()

    _YVector = type('_YVector', (meta.SDValueVector,), dict())
    _YMatrix = type('_YMatrix', (meta.SDValueMatrix,), dict())
    _YColor = type('_YColor', (meta.SDValueColorRGBA,), dict())

elif env.Nuke():
    import _nukemath

    _YVector = type('_YVector', (_nukemath.Vector3,), dict())
    _YMatrix = type('_YMatrix', (_nukemath.Matrix4,), dict())

elif env.Max():
    import MaxPlus

    _YVector = type('_YVector', (MaxPlus.Point3,), dict())
    _YMatrix = type('_YMatrix', (MaxPlus.Matrix3,), dict())
    _YColor = type('_YColor', (MaxPlus.Color,), dict())




import sys

# server
HOST = 'localhost'
PORT = 1089

# user
EGG = "/Users/shosumioka/Library/Application Support/JetBrains/Toolbox/apps/PyCharm-P/ch-0/181.5087.37/PyCharm.app/Contents/debug-eggs/pycharm-debug.egg"
PYD = "/Users/shosumioka/Library/Application/ Support/JetBrains/Toolbox/apps/PyCharm-P/ch-0/181.5087.37/PyCharm.app/Contents/helpers/pydev"
YUR = "/Users/shosumioka/Documents/Yurlungur"

if EGG not in sys.path:
    sys.path.append(EGG)
    sys.path.append(PYD)
    sys.path.append(YUR)

import yurlungur as yr
yr.YNode("Camera").location.set(yr.YVector(1, 1, 1).array())

import pydevd
try:
    # import hrpyc
    # hrpyc.start_server(port=PORT, use_thread=False)

    pydevd.settrace(HOST, port=PORT, stdoutToServer=True, stderrToServer=True)
    yr.log(yr.name)
except:
    pydevd.stoptrace()


# hou sever side
# import hrpyc
# connection, hou = hrpyc.import_remote_module()

# http://www.sidefx.com/docs/houdini/hom/rpc.html
# /Applications/Blender/blender.app/Contents/Resources/2.79/scripts/startup
# /Applications/Houdini/Houdini16.5.473/Frameworks/Houdini.framework/Versions/16.5.473/Resources/houdini/python2.7libs


# -*- coding: utf-8 -*-
import os
import sys
import inspect
import subprocess
import tempfile
import multiprocessing
import multiprocessing.process as process

import yurlungur as yr
from yurlungur.core import env

local = os.path.dirname(os.path.dirname(inspect.currentframe().f_code.co_filename))
__all__ = map(lambda x: x[0], inspect.getmembers(sys.modules[__name__], inspect.isclass))

try:
    from yurlungur.Qt.QtCore import *
    from yurlungur.Qt.QtGui import *
    from yurlungur.Qt.QtWidgets import *
    from yurlungur.Qt import __binding__

    # script editor
    sys.path.append(os.path.join(
        os.path.dirname(local), "scriptEditor", "pw_multiScriptEditor")
    )
    from pw_multiScriptEditor import scriptEditorClass

except ImportError as e:
    pass


def mayapy(pystr):
    assert os.path.getsize(env.MayaBin)

    exe = sys.executable
    multiprocessing.set_executable(
        os.path.join(os.path.dirname(exe), "mayapy")
    )

    process.ORIGINAL_DIR = os.path.join(
        os.path.dirname(exe),
        "../Python/Lib/site-packages"
    )
    po = multiprocessing.Pool(4)

    subprocess.call(
        "{0}/bin/mayapy -c \"{1};{2};{3}\"".format(
            env.MayaBin, "import maya.standalone;maya.standalone.initialize(name='python')",
            "import sys; sys.path.append('{0}');".format(local) + pystr, "maya.standalone.uninitialize()"
        )
    )


def hython(pystr):
    assert os.path.getsize(env.HoudiniBin)
    subprocess.call(
        "{0}/hython -c\"import sys; sys.path.append('{1}');{2}\"".format(env.HoudiniBin, local, pystr)
    )


def bpython(pystr):
    assert os.path.getsize(env.BlenderBin)
    subprocess.call(
        "{0}.blender --python-expr {1} -b".format(env.BlenderBin, pystr)
    )


def maxpy(pystr):
    # http://help.autodesk.com/view/3DSMAX/2019/ENU/?guid=GUID-96D3ABE3-32CA-491D-9CAD-0A0576346E54
    assert os.path.getsize(env.MaxBin)
    subprocess.call(
        "{0}/3dsmaxpy -c \"{1};{2}\"".format(env.MaxBin, sys.path.append(yr), pystr)
    )


def uepython(project, pystr):
    assert os.path.getsize(env.UnrealBin) or os.path.exists(project)

    # temp
    with tempfile.NamedTemporaryFile(delete=False) as tf:
        tf.write(pystr)
    subprocess.call(
        "{0}/UE4Editor-Cmd -run=pythonscript -script={1}".format(env.UnrealBin, pyfile)
    )


def _cli(args):
    """
    command line parser
    """
    try:
        import argparse
    except ImportError:
        yr.logger.warn("argparse is not found.")
        sys.exit(1)

    parser = argparse.ArgumentParser(
        prog='yr.cli',
        usage='Demonstration of argparser',
        description='yurlungur console.',
        epilog="{0} v.{1} {2}".format(yr.name, yr.version, sys.executable),
        add_help=True,
    )
    parser.add_argument("--dialog", "-d",
                        help="Launch yr.widget if Qt is installed.",
                        action="store_true")
    parser.add_argument("--mayapy", "-ma",
                        help="Run Python from mayapy.",
                        nargs=1)
    parser.add_argument("--hython", "-hou",
                        help="Run Python from hython.",
                        nargs=1)
    parser.add_argument("--unrealpy", "-ue",
                        help="Run Python from unreal editor cmd.",
                        nargs=2)
    parser.add_argument("--maxpy", "-max",
                        help="Run Python from 3dsmaxpy.",
                        nargs=2)

    args = parser.parse_args(args)
    if args.dialog:
        main()

    if args.mayapy:
        mayapy(args.mayapy[0])

    if args.hython:
        hython(args.hython[0])

    if args.unrealpy:
        project, expr = args.unrealpy
        uepython(project, expr)

    if args.maxpy:
        maxpy(args.maxpy[0])


def main(args=[]):
    if yr.Qt():
        app = QApplication(args)
        widget = scriptEditorClass()
        widget.show()
        sys.exit(app.exec_())
    else:
        yr.logger.warn("Qt isn't available")


if __name__ == '__main__':
    main(sys.argv[1:])



# -*- coding: utf-8 -*-
import sys
import inspect
import cmath

from yurlungur.core.wrapper import (
    _YVector, _YMatrix, _YColor
)
from yurlungur.core.env import Numpy, Blender

if Numpy():
    import numpy as np
else:
    pass


class YVector(_YVector):
    def __init__(self, *args, **kwargs):
        if Blender():
            super(YVector, self).__init__()
            self.vector = [self.x, self.y, self.z]
        else:
            super(YVector, self).__init__(*args, **kwargs)
            self.vector = args

    @Numpy
    def array(self):
        return np.array(self.vector, dtype=np.float16)

    def identify(self):
        return

    def dot(self, a, b, norm=False):
        if norm:  # 正規化オプション
            a = self.normalize(a)
            b = self.normalize(b)
        dot = (a[0] * b[0]) + (a[1] * b[1])
        return dot

    def cross(self, a, b):
        return

    @Numpy
    def normalize(self, a):
        length = self.length(a)
        return [a[0] / length, a[1] / length]

    def length(self):
        return cmath.sqrt(self.x ** 2 + self.y ** 2)


class YMatrix(_YMatrix):
    def __init__(self, *args, **kwargs):
        super(YMatrix, self).__init__(*args, **kwargs)


class YColor(_YColor):
    def __init__(self, *args, **kwargs):
        super(YColor, self).__init__(*args, **kwargs)


__all__ = map(lambda x: x[0], inspect.getmembers(sys.modules[__name__], inspect.isclass))



from setuptools import setup

text = """
DCC tool scripting is almost used Python, but these api isn't similarly anything.
If you make lightweight tool, need to remember each application manners.
Yurlungur is common interface which adapted each application for universal wrapper.
"""

setup(
    name='yurlungur',
    version="0.9.5",
    url='https://sho7noka.github.io/Yurlungur/',
    license='MIT',
    author='sho7noka',
    author_email='shosumioka@gmail.com',
    platforms="any",
    description='universal scripting environment with Python which Maya, Houdini and Blender.',
    long_description=text,
    keywords=['maya', '3d', 'graphics', 'Game', 'VFX', 'CG', 'houdini', "unreal"],

    packages=["yurlungur"],
    include_package_data=True,
    test_suite='test'
)



# -*- coding: utf-8 -*-
import sys
import os
import inspect

# settings
sys.dont_write_bytecode = True
local = os.path.dirname(os.path.dirname(inspect.currentframe().f_code.co_filename))
if not local in sys.path:
    sys.path.append(local)

if __name__ == "__main__":
    from yurlungur.tool.standalone import _cli
    _cli(sys.argv[1:])



# -*- coding: utf-8 -*-
import fnmatch
from yurlungur.core.proxy import YNode
from yurlungur.tool.meta import meta


class _NodeType(object):
    def __getattr__(self, item):
        if hasattr(meta, "types"):
            nodes = fnmatch.filter(dir(meta.types), str(item))
        else:
            nodes = self.findNodes(item)

        for node in nodes:
            setattr(self, str(item), YNode(node))

        return YNode(item)

    def findNodes(self, pattern):
        if hasattr(meta, "listNodeTypes"):
            # http://help.autodesk.com/cloudhelp/2016/JPN/Maya-Tech-Docs/CommandsPython/shadingNode.html
            categories = [
                "geometry", "camera", "light", "utility",
                "color", "shader", "texture", "rendering", "postprocess"
            ]
            # meta.allNodeTypes(ia=1)
            for category in categories:
                yield fnmatch.filter(meta.listNodeTypes(category), pattern)

        if hasattr(meta, "nodeTypes"):
            for category in meta.nodeTypeCategories().keys():
                yield fnmatch.filter(
                    meta.nodeTypeCategories()[category].nodeTypes().keys(),
                    pattern
                )


YType = _NodeType()



# coding: utf-8
import os
from yurlungur.tool.meta import meta


class Base(object):
    """
    projects/sequences/tracks/clips
    """

    def to_object(self, val):
        return YObject(val)

    def to_file(self, val):
        return YFile(val)


class Render(object):
    def __init__(self, project):
        self.project = project
        self.jobs = project.GetRenderJobs()
        self.presets = project.GetRenderPresets()
        self.formats = project.GetRenderFormats()
        # GetCurrentRenderFormatAndCodec

    def add(self):
        self.project.AddRenderJob()

    def start(self, *args):
        """args 無しも可能"""
        self.project.StartRendering(*args)

    def stop(self):
        if self.project.IsRenderingInProgress():
            self.project.StopRendering()

    def delete(self, *args):
        if len(args) == 0:
            self.project.DeleteAllRenderJobs()
        else:
            self.project.DeleteRenderJobIndex(args[0])


class Projects(object):

    def __init__(self):
        self.manager = meta.resolve.GetProjectManager()
        self.project = self.manager.GetCurrentProject()
        self.projects = self.manager.GetProjectsInCurrentFolder()

    def __repr__(self):
        return self.project.GetName()

    def __getitem__(self, val):
        if val in self.projects:
            self.project = self.manager.LoadProject(val)
        else:
            self.project = self.manager.CreateProject(val)

        return self

    @property
    def sequences(self):
        return Timeline(self.project)

    @property
    def render(self):
        return Render(self.project)


class Timeline(object):
    def __init__(self, project):
        self.project = project
        self.timeline = project.GetCurrentTimeline()
        self.media = project.GetMediaPool()
        self.clips = self.media.GetRootFolder().GetClips().values()

    def __repr__(self):
        return self.timeline.GetName() + self.timeline.GetCurrentTimecode()

    def __getitem__(self, val):
        assert isinstance(val, (str, int))

        if type(val) == int:
            if 0 < val < self.project.GetTimelineCount():
                self.timeline = self.project.GetTimelineByIndex(val)

        elif type(val) == str:
            if os.path.exists(val):
                self.timeline = self.media.ImportTimelineFromFile(val)
            elif val in [v.GetName() for v in self.clips]:
                self.project.SetCurrentTimeline(val)
                self.timeline = self.project.GetCurrentTimeline()
            else:
                self.timeline = self.media.CreateEmptyTimeline(val)

        return self

    def imports(self, *args):
        self.media.AppendToTimeline(*args)
        self.timeline = self.project.GetCurrentTimeline()
        return self

    @property
    def tracks(self):
        return Track(self.timeline)


class Track(object):
    def __init__(self, timeline):
        self.timeline = timeline
        self.track = timeline.GetCurrentVideoItem()

    def __repr__(self):
        return self.track.GetName()

    def __getitem__(self, val):
        assert isinstance(val, int)

        if 0 < val < self.timeline.GetTrackCount("video"):
            self.track = self.timeline.GetItemsInTrack("video", val)
        return self

    @property
    def clips(self):
        return Item(self.track)


class Item(object):
    def __init__(self, track):
        self.track = track

    def __repr__(self):
        return self.track.GetName() + self.track.GetDuration()

    def __getitem__(self, val):
        assert isinstance(val, (str, int))

        if type(val) == int:
            if 0 < val < self.track.GetFusionCompCount():
                self.track = self.track.GetFusionCompByIndex(val)

        elif type(val) == str:
            for v in self.track.GetFusionCompNames().values():
                if val in v.GetName():
                    self.track = self.track.GetFusionCompByName(val)
            if os.path.exists(val):
                self.track = self.track.ImportFusionComp(val)
                self.track = self.track.LoadFusionCompByName(val)
            else:
                self.track = self.track.AddFusionComp()
                self.track.RenameFusionCompByName(self.track.GetName(), val)

        return self

    def exports(self, path):
        return self.track.ExportFusionComp(path, 1)

    def delete(self, name):
        return self.track.DeleteFusionCompByName(name)

# clips
# clips = resolve.GetMediaStorage().AddItemsToMediaPool(paths)
# clips.GetClipProperty()
# clips.SetClipProperty(n, v)



from yurlungur.core.app import application
from yurlungur.core.command import cmd, file
from yurlungur.core.deco import UndoGroup
from yurlungur.core.env import Maya, Houdini, Blender
from yurlungur.core.logger import log
from yurlungur.core.nodetype import YType
from yurlungur.core.proxy import YObject, YFile, YNode



import doctest
import unittest
import sys

sys.path.append('../yurlungur')

import yurlungur as yr
from yurlungur.core.env import installed, Max
from yurlungur.tool import standalone


@unittest.skipUnless(installed("Max"), "Max is not found")
class TestMax(unittest.TestCase):
    def test_env(self):
        assert Max()

    def test_Max(self):
        standalone.maxpy("import yurlungur as yr; print yr.name")

    @unittest.skip("only runtime")
    def test_cmds(self):
        box = yr.YNode().create('box')
        print box.name, box.id, box.attr('Width'), box.attrs

        print box.create('Mirror', offset=0.1)
        box.width.set(10)
        box.pos.set(yr.YVector(1, 100, 1))
        box.select()
        box('aaa')
        box = yr.YNode('aaa')
        print box.geometry()
        box2 = box.instance()
        print box2.geometry()

        yr.meta.eval('$%s.width = 20' % box.name)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import os
import sys
import functools
import platform


def __import__(name, globals=None, locals=None, fromlist=None):
    # Fast path: see if the module has already been imported.
    try:
        return sys.modules[name]
    except KeyError:
        pass

    if "DaVinci" in name:
        if Windows():
            resolve = "%PROGRAMDATA%\\Blackmagic Design\\DaVinci Resolve\\Support\\Developer\\Scripting\\Modules"
        if MacOS():
            resolve = "/Library/Application Support/Blackmagic Design/DaVinci Resolve/Developer/Scripting/Modules"
        if Linux():
            resolve = "/opt/resolve/Developer/Scripting/Modules"
        sys.path.append(resolve)

    try:
        import imp
    except ImportError:
        from importlib import import_module
        return import_module(name)

    try:
        fp, pathname, description = imp.find_module(name)
        return imp.load_module(name, fp, pathname, description)
    except ImportError:
        return False


def Qt(func=None):
    try:
        import yurlungur.Qt as Qt
        is_Qt = any([Qt])
    except ImportError:
        return False

    if func is None:
        return is_Qt

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if is_Qt:
            return func(*args, **kwargs)

    return wrapper


def Numpy(func=None):
    try:
        import numpy as nm
        is_numpy = True
    except ImportError:
        return False

    if func is None:
        return is_numpy

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if is_numpy:
            return func(*args, **kwargs)

    return wrapper


def Windows(func=None):
    if func is None:
        return platform.system() == "Windows"

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if platform.system() == "Windows":
            return func(*args, **kwargs)

    return wrapper


def Linux(func=None):
    if func is None:
        return platform.system() == "Linux"

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if platform.system() == "Linux":
            return func(*args, **kwargs)

    return wrapper


def MacOS(func=None):
    if func is None:
        return platform.system() == "Darwin"

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if platform.system() == "Darwin":
            return func(*args, **kwargs)

    return wrapper


def Maya(func=None):
    if func is None:
        return "maya" in sys.executable

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if "maya" in sys.executable:
            return func(*args, **kwargs)

    return wrapper


def Houdini(func=None):
    if func is None:
        return __import__("hou")

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if __import__("hou"):
            return func(*args, **kwargs)

    return wrapper


def Unreal(func=None):
    if func is None:
        return "UE4" in sys.executable

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if "UE4" in sys.executable:
            return func(*args, **kwargs)

    return wrapper


def Unity(func=None):
    if func is None:
        return

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if "UE4" in sys.executable:
            return func(*args, **kwargs)

    return wrapper


def Blender(func=None):
    if func is None:
        return "blender" in sys.executable

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if "blender" in sys.executable:
            return func(*args, **kwargs)

    return wrapper


def Max(func=None):
    if func is None:
        return "3dsmax" in sys.executable

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if "3dsmax" in sys.executable:
            return func(*args, **kwargs)

    return wrapper


def Substance(func=None):
    if func is None:
        return "Substance" in sys.executable

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if "Substance" in sys.executable:
            return func(*args, **kwargs)

    return wrapper


def Davinci(func=None):
    if func is None:
        return __import__("DaVinciResolveScript")

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if __import__("DaVinciResolveScript"):
            return func(*args, **kwargs)

    return wrapper


def Nuke(func=None):
    if func is None:
        return "Nuke" in sys.executable

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if "Nuke" in sys.executable:
            return func(*args, **kwargs)

    return wrapper


def Gaffer(func=None):
    if func is None:
        return __import__("GafferScene")

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if __import__("GafferScene"):
            return func(*args, **kwargs)

    return wrapper


def installed(app):
    _app = app.lower()

    if _app == "maya":
        return os.path.exists(_Maya())
    if _app == "houdini":
        return os.path.exists(_Houdini())
    if _app == "unreal":
        return os.path.exists(_Unreal())
    if _app == "blender":
        return os.path.exists(_Blender())
    if _app == "max":
        return os.path.exists(_Max())
    if _app == "Substance":
        return os.path.exists(_Substance())
    return False


def _Maya():
    """find Maya app"""
    d = {
        "Linux": "/usr/autodesk/maya2017-x64",
        "Windows": "C:/Program Files/Autodesk/Maya2017",
        "Darwin": "/Applications/Autodesk/maya2017/Maya.app/Contents",
    }
    return os.environ.get("MAYA_LOCATION") or d[platform.system()]


def _Houdini():
    """find Houdini app"""
    d = {
        "Linux": "/usr/autodesk/maya2017-x64",
        "Windows": "C:/Program Files/Side Effects Software/Houdini 16.5.323/bin",
        "Darwin": "/Applications/Houdini/Houdini16.5.473/Houdini.app/Contents",
    }
    return os.environ.get("HIP") or d[platform.system()]


def _Substance():
    d = {
        "Linux": "/usr/autodesk/maya2017-x64",
        "Windows": "C:/Program Files/Side Effects Software/Houdini 16.5.323/bin",
        "Darwin": "/Applications/Substance Designer.app/Contents",
    }
    return d[platform.system()]


def _Max():
    """find 3dsMax app"""
    return os.environ.get("ADSK_3DSMAX_X64_2019") or "C:/Program Files/Autodesk/3ds Max 2019"


def _Blender():
    """find Blender app"""
    d = {
        "Linux": "",
        "Windows": "C:/Program Files/Blender Foundation/Blender",
        "Darwin": "/Applications/Blender/blender.app/Contents/MacOS/blender"
    }
    return d[platform.system()]


def _Unreal():
    d = {
        "Linux": "",
        "Windows": "C:/Program Files/Epic Games/UE_4.20/Engine/Binaries/Win64",
        "Darwin": ""
    }
    return d[platform.system()]


MayaBin = _Maya()
HoudiniBin = _Houdini()
BlenderBin = _Blender()
MaxBin = _Max()

__all__ = [
    "Windows", "Linux", "MacOS",
    "Maya", "Houdini", "Blender",
    "MayaBin", "HoudiniBin", "BlenderBin",
    "MaxBin"
]
